function [EleOut,cut,AS,AD,ws,wd,wmin,wmax] = graphcut(V,Ele,EleN,w,varargin)
  % GRAPHCUT Segment a elements E of a hypervolume-mesh based on winding number w
  %
  % [EleOut,cut] = graphcut(V,Ele,EleN,w)
  % 
  % Inputs:
  %   V   #V by dim list of vertices
  %   Ele  #Ele by dim+1 list of element indices
  %   EleN  #Ele by dim+1 list of element neighbors (see cdt.m or tt.m), -1
  %     means boundary facing
  %   w  #Ele list of winding number values. Normalized to
  %     [0->outside,1->inside]!!
  %   Optional:
  %     'SmoothnessWeight'  followed by weight on smoothness term
  %        larger (1e6) --> less splintery, but less faithful to data. So
  %          choose this when data is less trustworthy (i.e. more noisy)
  %        smaller {1e1} --> more faithful to local data, more splintery
  %      'DataWeight'  Followed by weight on data term {255}
  %      'Sigma' Followed by smoothness exponential's sigma as described in
  %        "Graph Cuts and Efficient N-D Image Segmentation" {1e0}
  %      'InterfaceAreaPower'  followed by power applied to interface
  %        hype-surface-areas weighting smoothness terms {1}
  %      'VolumeWeight'  followed by value between [0,1] balancing between
  %        completely volume-weighted data term and non-volume (uniformly)
  %        weighted data term {1}
  %      'SourceSinkValues'  followed by array of [ws,wd] the source and sink
  %        data values used to construct data term 
  %          OR
  %        string describing source and sink values method:
  %          'min-max'  use min and max of w (for shifted input)
  %          '0-1'  use 0 and 1 (for well bounded input)
  %          'k-means'  for arbitrary input, k=2 means cluster hoping that
  %             good values are found at cluster centers
  %        
  % Outputs:
  %   EleOut   #EleOut by dim+1 list of output element indices
  %   cut  #Ele list of cut assignments
  %   AS  #Ele by #Ele sparse matrix of smoothness weights
  %   AD  #Ele by 2 dense matrix of source and sink data weights
  % 

  % Default parameter values

  % smoothness term weight
  sm = 1e1;
  sigma = 1e0;
  % smoothness term Interface hypersurface area power
  int_p = 1;
  % data term weight
  da = 255;
  % volume weight
  pvol = 1;
  % method for determining source and sink values for data term
  source_sink_values = 'k-means';
  % Whose max-flow code to use. Doesn't seem to matter
  %max_flow_code = 'matlab_bgl';
  max_flow_code = 'kolmogorov';

  ii = 1;
  while ii < numel(varargin)
    switch varargin{ii}
    case 'SmoothnessWeight'
      assert(ii+1<=numel(varargin));
      ii = ii+1;
      sm = varargin{ii};
    case 'DataWeight'
      assert(ii+1<=numel(varargin));
      ii = ii+1;
      da = varargin{ii};
    case 'InterfaceAreaPower'
      assert(ii+1<=numel(varargin));
      ii = ii+1;
      int_p = varargin{ii};
    case 'VolumeWeight'
      assert(ii+1<=numel(varargin));
      ii = ii+1;
      pvol = varargin{ii};
    case 'MaxFlowCode'
      assert(ii+1<=numel(varargin));
      ii = ii+1;
      max_flow_code = varargin{ii};
    case 'Sigma'
      assert(ii+1<=numel(varargin));
      ii = ii+1;
      sigma = varargin{ii};
    case 'SourceSinkValues'
      assert(ii+1<=numel(varargin));
      ii = ii+1;
      source_sink_values = varargin{ii};
      if ~ischar(source_sink_values)
        assert(numel(source_sink_values)==2);
        ws = source_sink_values(1);
        wd = source_sink_values(2);
        source_sink_values = 'given';
      end
    otherwise
      error('Unsupported parameter: %s',varargin{ii});
    end
    ii = ii+1;
  end

  % To be nice, catch unnormalized w
  if sum(round((abs(w)-2*pi)*1e3)==0) > 10
    warning('w seems to be between [0,2*pi]. Should be between [0,1]');
  end
  if sum(round((abs(w)-4*pi)*1e3)==0) > 10
    warning('w seems to be between [0,4*pi]. Should be between [0,1]');
  end

  % Flip sign to match wikipedia convention (CCW)
  cw = mean(w)<0;
  if cw
    w = -1*w;
    warning('Flipped sign of w to match CCW convention');
  end

  % number of dimensions
  dim = size(Ele,2)-1;
  assert(dim == size(V,2));
  % number of elements
  m = size(Ele,1);
  % number of vertices
  n = size(V,1);

  % all edges between elements
  EleE = [ repmat(1:m,1,dim+1)' EleN(:)];
  % compute interface hyper-surface areas and element hypervolumes
  switch dim
  case 3
    % Interface triangles
    EleI = [Ele(:,[2 3 4]); Ele(:,[3 4 1]); Ele(:,[4 1 2]); Ele(:,[1 2 3])];
    % Only keep real interfaces
    EleI = EleI(EleE(:,2) ~= -1,:);
    % Interface hypersurface-areas
    lEleE = doublearea(V,EleI)/2+eps;
    % compute signed volumes
    vol = volume(V,Ele);
  case 2
    % Interface lengths
    l = [ ...
      sqrt(sum((V(Ele(:,2),:)-V(Ele(:,3),:)).^2,2)); ...
      sqrt(sum((V(Ele(:,3),:)-V(Ele(:,1),:)).^2,2)); ...
      sqrt(sum((V(Ele(:,1),:)-V(Ele(:,2),:)).^2,2)); ...
      ];
    % only keep real interfaces
    lEleE = l(EleE(:,2) ~= -1);
    % "Volumes" (signed areas)
    vol = doublearea(V,Ele)/2;
  otherwise
    error('Unsupported dimension (%d)',dim);
  end
  assert(all(lEleE>=0));
  if ~all(vol>=0)
     warning('Negative volumes found...');
     vol = abs(vol);
  end
  % normalize
  lEleE = lEleE./mean(lEleE);
  vol = vol./mean(vol);
  % only keep real edges
  EleE = EleE(EleE(:,2) ~= -1,:);

  % Edge-weights adjacency matrix for smoothness term and keep room for data
  % edge weights to source and sink
  A = sparse( ...
    EleE(:,1), ...
    EleE(:,2), ...
    sm.*lEleE.^int_p.*exp(-(abs(w(EleE(:,1))-w(EleE(:,2))).^2)/(2*sigma^2)), ...
    m+2, ...
    m+2);
  % Should be symmetric
  assert(normest(A-A')<sqrt(eps));

  % Set up auxillary source and sink
  s = m+1;
  d = m+2;

  % Set data values
  switch source_sink_values
  case 'min-max'
    ws = min(w);
    wd = max(w);
  case 'k-means'
    % Choose high and low, expecting these to be cluster centers.
    [~,c] = kmeans(w,2,'Start',[0;1]);
    ws = min(c);
    wd = max(c);
  case '0-1'
    ws = 0;
    wd = 1;
  case 'given'
    % do nothing, ws and wd are already set
  case 'spatially-varying'
    % Assume all vertices were on original surface

    % Q: What about vertices on the boundary (convex hull)?
    % A: They should have a phantom neighboring element with w=0.
    % Q: But what if the input is not closed? Is zero still correct?
    % A: no.
    % Q: What about distinquishing between boundary vertices and ear corners?
    % Gather boundary vertices
    switch dim
    case 2
      b = unique(reshape(outline(Ele),[],1));
    case 3
      b = unique(reshape(boundary_faces(Ele),[],1));
    end
    bmask = ismember(1:n,b)';

    % vertex to element adjacency matrix
    V2E = sparse(Ele(:),repmat(1:m,1,dim+1)',1,n,m);
    % Min and max w of incident elements
    wmin = full(minnz( ...
      sparse(Ele(:),repmat(1:m,1,dim+1)',repmat(w+(w==0)*eps,dim+1,1),n,m)')');
    wmax = full(maxnz( ...
      sparse(Ele(:),repmat(1:m,1,dim+1)',repmat(w+(w==0)*eps,dim+1,1),n,m)')');

    % Q: Can an interior vertex have wmin == wmax ?
    assert(~any(wmin==wmax & ~bmask));
    weq = wmin==wmax;
    % Zero is always a safe choice for "outside"
    % Q: What about wmax - 1?
    % A: Yes, that just assumes that everything the vertex touches must be inside.
    % Q: What about "outlier detection" then?
    % A: Could try something like: wmax/2 if wmax<1 ...
    wmin(weq) = 0;

    %% Simplest thing: average min and max from incident vertices
    %ws = mean(wmin(Ele),2);
    %wd = mean(wmax(Ele),2);
    %ws = min(wmin(Ele),[],2);
    %wd = max(wmax(Ele),[],2);

    % We're not trying to determin ``how inside an element is'', rather just
    % whether it is outside
    %ws = mean(wmin(Ele),2);
    %wd = min(ws + 1,mean(wmax(Ele),2));
    %ws = min(wmin(Ele),[],2);
    %wd = min(wmax(Ele),[],2);
    %min(wd-ws)
    ws = mean([min(wmin(Ele),[],2) min(wmax(Ele),[],2)],2)-1/2;
    wd = mean([min(wmin(Ele),[],2) min(wmax(Ele),[],2)],2)+1/2;

    %% A well predicated ws,wd expects to place w at 0.5
    %a = min(max(w-ws,0),1);
    %ws = (1-a)*0 + a.*ws;
    %wd = (1-a)*1 + a.*wd;

  otherwise
    error('Unsupported SourceSinkValues value: %s',source_sink_values);
  end

  A(s,1:m) = da*((1-pvol) + pvol*vol).*(max(w-ws,0)).^2;
  A(1:m,d) = da*((1-pvol) + pvol*vol).*(max(wd-w,0)).^2;

  % Smoothness portion
  AS = A(1:m,1:m);
  % Data portion
  AD = [A(1:m,d) A(s,1:m)'];

  switch max_flow_code
  case 'matlab_bgl'
    options.algname = 'push_relabel';
    old_warn = warning('OFF','MATLAB:dispatcher:pathWarning');
    addpath matlab_bgl/
    warning(old_warn);
    % This code requires integer values (stored in double matrices :-)
    A = round(A);
    AS = round(AS);
    AD = round(AD);
    tic;
    [flow cut RG FG] =  max_flow(A,s,d,options);
    toc
  case 'kolmogorov'
    old_warn = warning('OFF','MATLAB:dispatcher:pathWarning');
    addpath maxflow/
    warning(old_warn);
    tic;
    [flow cut] =  maxflow(AS,AD);
    toc
    %cut = double(cut);
    %[AI,AJ,AV] = find(AS);
    %CG = sparse(AI,AJ,cut(AI)~=cut(AJ),size(AS,1),size(AS,2));
  end

  % throw away souce and dest
  cut = cut(1:size(w,1));
  cut = cut==1;

  % CUT!
  EleOut = Ele(cut,:);


end
